/*****************************************************************************/
/*
                                  AUTHVMS.c


    THE GNU GENERAL PUBLIC LICENSE APPLIES DOUBLY TO ANYTHING TO DO WITH
                    AUTHENTICATION AND AUTHORIZATION!

    This package is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; version 2 of the License, or any later
    version.

>   This package is distributed in the hope that it will be useful,
>   but WITHOUT ANY WARRANTY; without even the implied warranty of
>   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
>   GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


ACME AUTHENTICATOR
------------------
The Authentication and Credentials Management Extensions (ACME) subsystem
provides authentication and persona-based credential services.  Applications
use these services to enforce authentication policies defined by ACME agents
running in the context of the ACME_SERVER process.

It is not available on VAX and is only for VMS later than V7.3.

Relies on the image being INSTALLed with /AUTHPRIVILEGE=(IMPERSONATE) so that
it can use $ACMW.


AUTHVMS AUTHENTICATOR
--------------------
On platforms that do not support ACME a composite authenticator is provided
using $SCAN_INTRUSION and $GETUAI.  This implementation checks for disusered
flag, etc., primary password, password expiry (primary and secondary).  It also
checks and maintains the intrusion database.  Every effort has been made to
ensure this authenticator is robust but no guarantees.  I'd much rather rely on
an Engineering maintained service such as $ACM!

Relies on the image being INSTALLed with /AUTHPRIVILEGE=(SECURITY) so that it
can use $SCAN_INTRUSION.


COPYRIGHT
---------
Copyright (C) 2008-2013 Mark G.Daniel
This program, comes with ABSOLUTELY NO WARRANTY.
This is free software, and you are welcome to redistribute it under the
conditions of the GNU GENERAL PUBLIC LICENSE, version 3, or any later version.
http://www.gnu.org/licenses/gpl.txt


VERSION HISTORY
---------------
05-MAY-2009  MDG  AuthVmsVerifyUser() only employ ACME$_TARGET_DOI_NAME
                    when configured otherwise ACME$LATEST_ENABLED_AGENT_LIST
11-APR-2009  MGD  AuthVmsCheckIdent()
24-MAR-2008  MGD  development commences
*/
/*****************************************************************************/

#ifndef AUTHVMS_ACME
#  ifndef __VAX
      /* only available for V7.3 and later */
#     if (__CRTL_VER >= 70300000)
#        define AUTHVMS_ACME 1
#     else
#        define AUTHVMS_ACME 0
#     endif
#  else
#     define AUTHVMS_ACME 0
#  endif
#endif

#ifdef WASD_VMS_V6
#undef _VMS_V6_SOURCE
#define _VMS_V6_SOURCE
#undef __VMS_VER
#define __VMS_VER 60200000
#undef __CRTL_VER
#define __CRTL_VER 60200000
#endif

#ifndef __VAX
#   pragma nomember_alignment
#endif

/* standard C header files */
#include <ctype.h>
#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

/* VMS related header files */
#include <descrip.h>
#include <jpidef.h>
#include <libdef.h>
#include <libdtdef.h>
#include <prvdef.h>
#include <ssdef.h>
#include <stsdef.h>
#include <uaidef.h>

#ifndef UAI$M_PWDMIX
#define UAI$M_PWDMIX 0x2000000
#endif

/* this image is built with ACME capabilities */
#if AUTHVMS_ACME 
#include <utcblkdef.h>
#include <acmedef.h>
#include <acmemsgdef.h>
#ifndef LGI$_RESTRICT
#  define LGI$_RESTRICT 13861020
#endif
#define ILE3 ile3  /* because of the __VMS_VER */
#endif

/* application related header files */
#include "authvms.h"
#include "config.h"
#include "other.h"

#define INTRU_FACILITY "VWcms:"

#define FI_LI "AUTHVMS", __LINE__

/******************/
/* global storage */
/******************/

/********************/
/* external storage */
/********************/

extern BOOL  Debug,
             WatchEnabled;

extern CONFIG_DATA  VWcmsConfig;

/***********************/
/* required prototypes */
/***********************/

int lib$add_times (__unknown_params);
int sys$asctoid (__unknown_params);
int lib$sub_times (__unknown_params);
int sys$acmw (__unknown_params);
int sys$fao (__unknown_params);
int sys$find_held (__unknown_params);
int sys$gettim (__unknown_params);
int sys$getuai (__unknown_params);
int sys$hash_password (__unknown_params);
int sys$setprv (__unknown_params);
int sys$scan_intrusion (__unknown_params);

/*****************************************************************************/

/***************/
#if AUTHVMS_ACME 
/***************/

/*****************************************************************************/
/*
Verify username/password using $ACM services.
*/ 

int AuthVmsVerifyUser
(
char *UserName,
char *UserPassword,
char *RemoteHostName
)
{
   static unsigned long  DetachMask [2] = { PRV$M_DETACH, 0 };
   static unsigned long  AcmeLogonType = ACME$K_NETWORK;

   int  status,
        SetPrvStatus,
        UserNameLength;
   unsigned short  MappingAcmeNameLength,
                   MappedVmsUserNameLength;
   char  *cptr, *sptr, *zptr;
   char  MappedVmsUserName [32] = "",
         MappingAcmeName [32+1] = "",
         UserNameBuffer [256];
   ILE3  *itmptr;
   ILE3  AcmItemList [16];

   /* the SYS$ACM() IO status block is a little different to the norm! */
   unsigned long  AcmIOsb [4];

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis (FI_LI, "VERIFY ACME \"!AZ\"", UserName);

   itmptr = AcmItemList;

   if (VWcmsConfig.AuthVmsAcmeDoi)
   {
      itmptr->ile3$w_code = ACME$_TARGET_DOI_NAME;
      itmptr->ile3$w_length = strlen(VWcmsConfig.AuthVmsAcmeDoi);
      itmptr->ile3$ps_bufaddr = VWcmsConfig.AuthVmsAcmeDoi;
      itmptr->ile3$ps_retlen_addr = 0;
      itmptr++;
   }

   /* requires IMPERSONATE (DETACH) privilege on VMS V7.3-1 or earlier */
   itmptr->ile3$w_code = ACME$_LOGON_TYPE;
   itmptr->ile3$w_length = sizeof(AcmeLogonType);
   itmptr->ile3$ps_bufaddr = &AcmeLogonType;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_code = ACME$_PRINCIPAL_NAME_IN;
   itmptr->ile3$w_length = strlen(UserName);
   itmptr->ile3$ps_bufaddr = UserName;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   /* requires IMPERSONATE (DETACH) privilege */
   itmptr->ile3$w_code = ACME$_REMOTE_HOST_NAME;
   itmptr->ile3$w_length = strlen(RemoteHostName);
   itmptr->ile3$ps_bufaddr = RemoteHostName;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   /* create a composite string containing the user name */
   zptr = (sptr = UserNameBuffer) + sizeof(UserNameBuffer)-1;
   for (cptr = INTRU_FACILITY; *cptr; *sptr++ = *cptr++);
   for (cptr = UserName; *cptr && sptr < zptr; *sptr++ = toupper(*cptr++));
   *sptr = '\0';
   UserNameLength = sptr - UserNameBuffer;

   /* requires IMPERSONATE (DETACH) privilege */
   itmptr->ile3$w_code = ACME$_REMOTE_USERNAME;
   itmptr->ile3$w_length = UserNameLength;
   itmptr->ile3$ps_bufaddr = UserNameBuffer;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_code = ACME$_PASSWORD_1;
   itmptr->ile3$w_length = strlen(UserPassword);
   itmptr->ile3$ps_bufaddr = UserPassword;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_code = ACME$_MAPPED_VMS_USERNAME;
   itmptr->ile3$w_length = sizeof(MappedVmsUserName);
   itmptr->ile3$ps_bufaddr = MappedVmsUserName;
   itmptr->ile3$ps_retlen_addr = &MappedVmsUserNameLength;
   itmptr++;

   if (WatchEnabled)
   {
      itmptr->ile3$w_code = ACME$_MAPPING_ACME_NAME;
      itmptr->ile3$w_length = sizeof(MappingAcmeName);
      itmptr->ile3$ps_bufaddr = MappingAcmeName;
      itmptr->ile3$ps_retlen_addr = &MappingAcmeNameLength;
      itmptr++;
   }

   /* empty item to terminate the list */
   memset (itmptr, 0, sizeof(ILE3));

   /* provide the IMPERSONATE privilege */
   if (VMSnok (SetPrvStatus = sys$setprv (1, &DetachMask, 0, 0)))
      ErrorExit (SetPrvStatus, FI_LI);

   status = sys$acmw (0,
                      ACME$_FC_AUTHENTICATE_PRINCIPAL,
                      0,
                      &AcmItemList,
                      &AcmIOsb,
                      0,
                      0);

   if (WatchEnabled) WatchThis (FI_LI, "$ACMW %X!8XL", status);
   if (VMSnok (status)) ErrorExit (status, FI_LI);

   if (VMSnok (SetPrvStatus = sys$setprv (0, &DetachMask, 0, 0)))
      ErrorExit (SetPrvStatus, FI_LI);

   if (VMSok (AcmIOsb[0]))
   {
      MappedVmsUserName[MappedVmsUserNameLength] = '\0';
      if (WatchEnabled) MappingAcmeName[MappingAcmeNameLength] = '\0';
      /* going to assume storage is large enough here */
      strcpy (UserName, MappedVmsUserName);
   }
   else
      MappingAcmeName[0] = MappedVmsUserName[0] = '\0';

   if (WatchEnabled)
   {
      if (VMSok(AcmIOsb[0]) && AcmIOsb[0] == AcmIOsb[1])
         WatchThis (FI_LI, "ACME doi:\"!AZ\" agent:\"!AZ\" mapped:\"!AZ\" %X!8XL",
                    VWcmsConfig.AuthVmsAcmeDoi ?
                       VWcmsConfig.AuthVmsAcmeDoi : "*",
                    MappingAcmeName, MappedVmsUserName, AcmIOsb[0]);
      else
      if (!AcmIOsb[3] && AcmIOsb[0] == AcmIOsb[1])
         WatchThis (FI_LI, "ACME doi:\"!AZ\" sts:%X!8XL",
                    VWcmsConfig.AuthVmsAcmeDoi ?
                       VWcmsConfig.AuthVmsAcmeDoi : "*",
                    AcmIOsb[0]);
      else
         WatchThis (FI_LI, "ACME doi:\"!AZ\" sts:%X!8XL "
                    "sec:%X!8XL id:!UL acme:%X!8XL mapped:\"!AZ\"",
                    VWcmsConfig.AuthVmsAcmeDoi ?
                       VWcmsConfig.AuthVmsAcmeDoi : "*",
                    AcmIOsb[0], AcmIOsb[1],
                    AcmIOsb[2], AcmIOsb[3],
                    MappedVmsUserName);
   }

   if (AcmIOsb[0] == LGI$_RESTRICT)
   {
      /* authentication OK but restriction on login source */
      if (VWcmsConfig.AuthVmsAcmeNoRestrict)
      {
         if (WatchEnabled)
         WatchThis (FI_LI, "LGI$_RESTRICT - not applying source restriction");
         AcmIOsb[0] = SS$_NORMAL;
      }
   }

   return (AcmIOsb[0]);
}

/*****************************************************************************/

/***********************/
#else  /* AUTHVMS_ACME */
/***********************/

/*****************************************************************************/
/*
Verify username/password using $SCAN_INTRUSION and $GETUAI services.
*/ 

int AuthVmsVerifyUser
(
char *UserName,
char *UserPassword,
char *RemoteHostName
)
{
/* according to the documentation these can be returned by $SCAN_INTRUSION */
#define SECSRV$_INTRUDER 116299715
#define SECSRV$_SUSPECT 116299723
#define SECSRV$_SERVERNOTACTIVE 116311802
#define SECSRV$_NOSUCHINTRUDER 116311738 /* NOMATCH */
#define SECSRV$_INSUFINFO 116311754

/* just for convenience sake */
#define LOGIN_PRV$M_SECURITY 0x40

   static unsigned long  SecurityMask [2] = { 0, LOGIN_PRV$M_SECURITY };

   int  status,
        GetUaiStatus,
        SetPrvStatus;

   char  *cptr, *sptr, *zptr;

   char  UserNameBuffer [32+1];

   $DESCRIPTOR (UserNameDsc, UserNameBuffer);
   $DESCRIPTOR (RemoteHostNameDsc, RemoteHostName);

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis (FI_LI, "VERIFY AUTHVMS \"!AZ\"", UserName);

   /* create a composite string containing the user name */
   zptr = (sptr = UserNameBuffer) + sizeof(UserNameBuffer)-1;
   for (cptr = INTRU_FACILITY; *cptr; *sptr++ = *cptr++);
   for (cptr = UserName; *cptr && sptr < zptr; *sptr++ = toupper(*cptr++));
   *sptr = '\0';
   UserNameDsc.dsc$w_length = sptr - UserNameBuffer;

   RemoteHostNameDsc.dsc$a_pointer = RemoteHostName;
   RemoteHostNameDsc.dsc$w_length = strlen(RemoteHostName);

   /* provide the SECURITY privilege */
   if (VMSnok (SetPrvStatus = sys$setprv (1, &SecurityMask, 0, 0)))
      ErrorExit (SetPrvStatus, FI_LI);

   status = sys$scan_intrusion (SS$_NORMAL, &UserNameDsc, JPI$K_NETWORK,
                                0, &RemoteHostNameDsc, &UserNameDsc,
                                0, 0, 0, 0, 0);

   if (WatchEnabled) WatchThis (FI_LI, "$SCAN_INTRUSION %X!8XL", status);
   if (VMSnok (status)) ErrorExit (status, FI_LI);

   if (VMSnok (SetPrvStatus = sys$setprv (0, &SecurityMask, 0, 0)))
      ErrorExit (SetPrvStatus, FI_LI);

   if (status == SECSRV$_INTRUDER) return (SS$_INVLOGIN);
   if (status == SECSRV$_SERVERNOTACTIVE) return (SS$_INVLOGIN);
   if (status == SECSRV$_INSUFINFO) return (SS$_INVLOGIN);

   GetUaiStatus = AuthVmsGetUaiUser (UserName, UserPassword);
   if (WatchEnabled) WatchThis (FI_LI, "%X!8XL", GetUaiStatus);

   if (VMSnok (SetPrvStatus = sys$setprv (1, &SecurityMask, 0, 0)))
      ErrorExit (SetPrvStatus, FI_LI);

   status = sys$scan_intrusion (GetUaiStatus, &UserNameDsc, JPI$K_NETWORK,
                                0, &RemoteHostNameDsc, &UserNameDsc,
                                0, 0, 0, 0, 0);

   if (WatchEnabled) WatchThis (FI_LI, "$SCAN_INTRUSION %X!8XL", status);

   if (VMSnok (SetPrvStatus = sys$setprv (0, &SecurityMask, 0, 0)))
      ErrorExit (SetPrvStatus, FI_LI);

   return (GetUaiStatus);
}

/*****************************************************************************/
/*
Use $GETUAI and $HASH_PASSWORD to verify the user's credentials.  Also ensure
the account is not disusered, etc., and that the password(s) have not expired.
*/ 

int AuthVmsGetUaiUser
(
char *UserName,
char *UserPassword
)
{
   /* UAI flags that disallow access */
   static unsigned long  DisallowFlags =
          UAI$M_DISACNT | UAI$M_PWD_EXPIRED | UAI$M_PWD2_EXPIRED;

   int  status;

   unsigned long  Context,
                  UaiFlags,
                  UaiPrimeDays,
                  UaiNetworkAccessP,
                  UaiNetworkAccessS,
                  UaiRemoteAccessP,
                  UaiRemoteAccessS,
                  UaiUic;

   unsigned long  CurrentBinTime [2],
                  HashedPwd [2],
                  PwdExpBinTime [2],
                  ResultBinTime [2],
                  UaiExpBinTime [2],
                  UaiPwd [2],
                  UaiPwd2 [2],
                  UaiPwdDateBinTime [2],
                  UaiPwd2DateBinTime [2],
                  UaiPwdLifeBinTime [2];

   unsigned short  UaiSalt;

   unsigned char  UaiEncrypt;

   char  *cptr, *sptr, *zptr;

   char  PasswordBuffer [64+1],
         UserNameBuffer [15+1];

   $DESCRIPTOR (PasswordDsc, PasswordBuffer);
   $DESCRIPTOR (UserNameDsc, UserNameBuffer);

   ILE3  *itmptr;
   ILE3  UaiItemList [16];

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled) WatchThis (FI_LI, "GETUAI \"!AZ\"", UserName);

   zptr = (sptr = UserNameBuffer) + sizeof(UserNameBuffer)-1;
   for (cptr = UserName; *cptr && sptr < zptr; *sptr++ = toupper(*cptr++));
   *sptr = '\0';
   UserNameDsc.dsc$w_length = sptr - UserNameBuffer;

   itmptr = UaiItemList;

   itmptr->ile3$w_length = sizeof(UaiUic);
   itmptr->ile3$w_code = UAI$_UIC;
   itmptr->ile3$ps_bufaddr = &UaiUic;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiExpBinTime);
   itmptr->ile3$w_code = UAI$_EXPIRATION;
   itmptr->ile3$ps_bufaddr = &UaiExpBinTime;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiPwdDateBinTime);
   itmptr->ile3$w_code = UAI$_PWD_DATE;
   itmptr->ile3$ps_bufaddr = &UaiPwdDateBinTime;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiPwd2DateBinTime);
   itmptr->ile3$w_code = UAI$_PWD2_DATE;
   itmptr->ile3$ps_bufaddr = &UaiPwd2DateBinTime;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiPwdLifeBinTime);
   itmptr->ile3$w_code = UAI$_PWD_LIFETIME;
   itmptr->ile3$ps_bufaddr = &UaiPwdLifeBinTime;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiFlags);
   itmptr->ile3$w_code = UAI$_FLAGS;
   itmptr->ile3$ps_bufaddr = &UaiFlags;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiPwd);
   itmptr->ile3$w_code = UAI$_PWD;
   itmptr->ile3$ps_bufaddr = &UaiPwd;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiPwd2);
   itmptr->ile3$w_code = UAI$_PWD2;
   itmptr->ile3$ps_bufaddr = &UaiPwd2;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiEncrypt);
   itmptr->ile3$w_code = UAI$_ENCRYPT;
   itmptr->ile3$ps_bufaddr = &UaiEncrypt;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(UaiSalt);
   itmptr->ile3$w_code = UAI$_SALT;
   itmptr->ile3$ps_bufaddr = &UaiSalt;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = sizeof(unsigned long);
   itmptr->ile3$w_code = UAI$_PRIMEDAYS;
   itmptr->ile3$ps_bufaddr = &UaiPrimeDays;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = 3;
   itmptr->ile3$w_code = UAI$_NETWORK_ACCESS_P;
   itmptr->ile3$ps_bufaddr = &UaiNetworkAccessP;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = 3;
   itmptr->ile3$w_code = UAI$_NETWORK_ACCESS_S;
   itmptr->ile3$ps_bufaddr = &UaiNetworkAccessS;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = 3;
   itmptr->ile3$w_code = UAI$_REMOTE_ACCESS_P;
   itmptr->ile3$ps_bufaddr = &UaiRemoteAccessP;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   itmptr->ile3$w_length = 3;
   itmptr->ile3$w_code = UAI$_REMOTE_ACCESS_S;
   itmptr->ile3$ps_bufaddr = &UaiRemoteAccessS;
   itmptr->ile3$ps_retlen_addr = 0;
   itmptr++;

   memset (itmptr, 0, sizeof(ILE3));

   Context = -1;
   status = sys$getuai (0, &Context, &UserNameDsc, &UaiItemList, 0, 0, 0);

   if (VMSnok (status)) 
   {
      if (WatchEnabled) WatchThis (FI_LI, "$GETUAI %X!8XL", status);
      return (status);
   }

   if (WatchEnabled)
      WatchThis (FI_LI, "expire:!%D pwd:!%D pwd2:!%D life:!%D \
flags:!8XL prime:!8XL netp:!6XL nets:!6XL remp:!6XL rems:!6XL",
            &UaiExpBinTime,
            &UaiPwdDateBinTime,
            &UaiPwd2DateBinTime,
            &UaiPwdLifeBinTime,
            UaiFlags,
            UaiPrimeDays,
            UaiNetworkAccessP, UaiNetworkAccessS,
            UaiRemoteAccessP, UaiRemoteAccessS);

   zptr = (sptr = PasswordBuffer) + sizeof(PasswordBuffer)-1;
   if (UaiFlags & UAI$M_PWDMIX)
   {
      /* 7.3-2 mixed-case plus printable char passwords */
      for (cptr = UserPassword; *cptr && sptr < zptr; *sptr++ = *cptr++);
   }
   else
   {
      /* to uppercase! */
      for (cptr = UserPassword;
           *cptr && sptr < zptr;
           *sptr++ = toupper(*cptr++));
   }
   *sptr = '\0';
   PasswordDsc.dsc$w_length = sptr - PasswordBuffer;

   status = sys$hash_password (&PasswordDsc, UaiEncrypt,
                               UaiSalt, &UserNameDsc, &HashedPwd);
   if (VMSnok (status)) 
   {
      if (WatchEnabled) WatchThis (FI_LI, "$HASH_PASSWORD %X!8XL", status);
      return (status);
   }

   if (HashedPwd[0] != UaiPwd[0] ||
       HashedPwd[1] != UaiPwd[1])
   {
      if (WatchEnabled) WatchThis (FI_LI, "PASSWORD match failed");
      return (SS$_INVLOGIN);
   }

   /* automatically disallow if any of these flags are set! */
   if (UaiFlags & DisallowFlags)
   {
      if (WatchEnabled)
      {
         if (UaiFlags & UAI$M_DISACNT)
            cptr = "DISACNT";
         else
         if (UaiFlags & UAI$M_PWD_EXPIRED)
            cptr = "PWD_EXPIRED";
         else
         if (UaiFlags & UAI$M_PWD2_EXPIRED)
            cptr = "PWD2_EXPIRED";
         else
            cptr = "?";
         WatchThis (FI_LI, "FLAG failure !AZ", cptr);
      }
      return (SS$_INVLOGIN);
   }

   return (SS$_NORMAL);
}

/*****************************************************************************/

/************************/
#endif  /* AUTHVMS_ACME */
/************************/

/*****************************************************************************/
/*
Check whether the supplied username hold the supplied identifer.
Cache the most recent results just in case it's needed again.
*/ 

BOOL AuthVmsCheckIdent
(
char *UserName,
char *IdentName
)
{
   static unsigned long  IdentHeld,
                         IdentValue,
                         UaiUic;
   static char  UserNameBuffer [15+1],
                IdentNameBuffer [63+1];
   static $DESCRIPTOR (IdentNameDsc, "");
   static $DESCRIPTOR (UserNameDsc, UserNameBuffer);

   int  status;
   unsigned long  Context;
   unsigned long  FindHeldUic [2];
   char  *cptr, *sptr, *zptr;
   ILE3  *itmptr;
   ILE3  UaiItemList [2];

   /*********/
   /* begin */
   /*********/

   if (WatchEnabled)
       WatchThis (FI_LI, "RIGHTSID \"!AZ\" \"!AZ\"", UserName, IdentName);

   /* obviously if not authenticated against VMS then rights IDs impossible */
   if (!VWcmsConfig.LoginVms) return (FALSE);

   if (!(UserNameBuffer[0] && strsame (UserName, UserNameBuffer, -1)))
   {
      zptr = (sptr = UserNameBuffer) + sizeof(UserNameBuffer)-1;
      for (cptr = UserName; *cptr && sptr < zptr; *sptr++ = toupper(*cptr++));
      *sptr = '\0';
      UserNameDsc.dsc$w_length = sptr - UserNameBuffer;

      itmptr = UaiItemList;

      itmptr->ile3$w_length = sizeof(UaiUic);
      itmptr->ile3$w_code = UAI$_UIC;
      itmptr->ile3$ps_bufaddr = &UaiUic;
      itmptr->ile3$ps_retlen_addr = 0;
      itmptr++;

      memset (itmptr, 0, sizeof(ILE3));

      Context = -1;
      status = sys$getuai (0, &Context, &UserNameDsc, &UaiItemList, 0, 0, 0);

      if (VMSnok (status)) 
      {
         if (WatchEnabled) WatchThis (FI_LI, "$GETUAI %X!8XL", status);
         return (FALSE);
      }

      /* don't use any cached value */
      IdentHeld = 0;
   }

   if (!(IdentNameBuffer[0] && strsame (IdentName, IdentNameBuffer, -1)))
   {
      IdentNameDsc.dsc$a_pointer = IdentName;
      IdentNameDsc.dsc$w_length = strlen(IdentName);

      if (VMSnok (status = sys$asctoid (&IdentNameDsc, &IdentValue, 0))) 
      {
         if (WatchEnabled) WatchThis (FI_LI, "$ASCTOID %X!8XL", status);
         return (FALSE);
      }
   }

   /* if cached data matched then  */
   if (IdentValue && IdentHeld && IdentValue == IdentHeld) return (TRUE);

   Context = FindHeldUic[1] = 0;
   FindHeldUic[0] = UaiUic;

   while (VMSok (status =
          sys$find_held (&FindHeldUic, &IdentHeld, 0, &Context)))
      if (IdentHeld == IdentValue) break;

   if (VMSnok(status)) return (FALSE);

   return (TRUE);
}

/*****************************************************************************/

